package entities

import (
	"math/rand"
	"sort"

	tl "github.com/JoelOtter/termloop"
	"github.com/zladovan/gorched/core"
	"github.com/zladovan/gorched/debug"
	"github.com/zladovan/gorched/entities/terrain"
	"github.com/zladovan/gorched/physics"
)

// World represents game world with all entities.
// It extends from termloop.BaseLevel so it can be added to the screen as termloop.Level.
type World struct {
	*tl.BaseLevel
	terrain *terrain.Terrain
	physics *physics.Physics
	options WorldOptions
	// entitiesToRemove holds references to entities which will be removed on next Tick
	entitiesToRemove []tl.Drawable
	onEntityRemove   map[tl.Drawable]func()
}

// WorldOptions provide configuration needed for generating game world (one round).
type WorldOptions struct {
	// Width of game world in number of console pixels (cells)
	Width int
	// Height of game world in number of console pixels (cells)
	Height int
	// Seed is number used as random seed and if it is reused it allows to create same game looking game world with same positions for players
	Seed int64
	// AsciiOnly identifies that only ASCII characters can be used for world graphics
	ASCIIOnly bool
	// LowColor identifies that only 8 colors can be used for world graphics
	LowColor bool
}

// NewWorld creates new game world with all entities
func NewWorld(game core.Game, o WorldOptions) *World {
	// random positions in the world are seeded too
	rnd := rand.New(rand.NewSource(o.Seed))

	// create terrain
	terrain := terrain.Generate(&terrain.Generator{
		Seed:      o.Seed,
		Width:     o.Width,
		Height:    o.Height,
		Roughness: 7.5,
		LowColor:  o.LowColor,
	})

	// create clouds
	clouds := GenerateClouds(&CloudsGenerator{
		Seed:      o.Seed,
		Width:     o.Width,
		Height:    o.Height,
		LowColor:  o.LowColor,
		ASCIIOnly: o.ASCIIOnly,
	})

	// create players
	// TODO: update for different player counts
	tanks := []*Tank{
		NewTank(
			game.Players()[0],
			terrain.PositionOn(10+rnd.Intn(10)),
			0,
			tl.ColorRed,
			o.ASCIIOnly,
		),
		NewTank(
			game.Players()[1],
			terrain.PositionOn(o.Width-10-rnd.Intn(10)),
			180,
			tl.ColorBlack,
			o.ASCIIOnly,
		),
	}

	// cut the terrain around the tanks
	for _, tank := range tanks {
		x, y := tank.Position()
		w, h := tank.Size()
		terrain.CutAround(x, y+h, w)
	}

	// create trees
	trees := GenerateWood(&WoodGenerator{
		Line:      terrain.Line(),
		Seed:      o.Seed,
		Density:   0.2,
		MaxSize:   6,
		MinSpace:  1,
		LowColor:  o.LowColor,
		ASCIIOnly: o.ASCIIOnly,
	})

	// cut the trees around the tanks
	for _, tank := range tanks {
		x, y := tank.Position()
		w, h := tank.Size()
		trees = trees.CutAround(x, y, w, h)
	}

	// create level with all entities
	bg := tl.Attr(111)
	if o.LowColor {
		bg = tl.ColorBlue
	}
	world := &World{
		BaseLevel:      tl.NewBaseLevel(tl.Cell{Bg: bg}),
		terrain:        terrain,
		physics:        &physics.Physics{Gravity: 9.81, Ground: terrain.HeightInside},
		options:        o,
		onEntityRemove: map[tl.Drawable]func(){},
	}
	world.AddEntity(clouds)
	for _, c := range terrain.Entities() {
		world.AddEntity(c)
	}
	for _, t := range trees {
		world.AddEntity(t)
	}
	for _, t := range tanks {
		world.AddEntity(t)
	}

	debug.Logf("New world created width=%d height=%d seed=%d", o.Width, o.Height, o.Seed)

	return world
}

// RemoveEntity only registers entity to remove.
// Entity will be removed in next Tick.
// This is needed for be able to remove entities from Draw method (where level is accessible).
func (w *World) RemoveEntity(e tl.Drawable) {
	w.entitiesToRemove = append(w.entitiesToRemove, e)
}

// Draw draws all entities in the world.
// Drawing takes into account z-index of entity which can be specified by implementing ZIndexer interface.
// Gravity is also applied here as there is access to delta time from screen.
func (w *World) Draw(s *tl.Screen) {
	// depthField  will contain entities distributed by z-index
	depthField := make(map[int][]tl.Drawable, len(w.Entities))

	for _, e := range w.BaseLevel.Entities {

		// apply physics to all entities with bodies
		if entity, ok := e.(physics.HasBody); ok {
			w.physics.Apply(entity, s.TimeDelta())
		}

		// find z-index of entity where 0 is the default
		zIndex := 0
		if entity, ok := e.(ZIndexer); ok {
			zIndex = entity.ZIndex()
		}

		// add entity to depth field
		depthField[zIndex] = append(depthField[zIndex], e)
	}

	// find all different values of z-index and sort them
	zIndexes := []int{}
	for z := range depthField {
		zIndexes = append(zIndexes, z)
	}
	sort.Ints(zIndexes)

	// draw entities in order defined by z-index (lower first)
	for _, z := range zIndexes {
		for _, e := range depthField[z] {
			e.Draw(s)
		}
	}
}

// Tick first removes all entity previously registered to be removed.
// Then calls original Tick logic.
func (w *World) Tick(e tl.Event) {
	for _, entity := range w.entitiesToRemove {
		w.BaseLevel.RemoveEntity(entity)
		if callback := w.onEntityRemove[entity]; callback != nil {
			callback()
			w.onEntityRemove[entity] = nil
		}
	}
	w.BaseLevel.Tick(e)
}

// OnEntityRemove registers callback f which will be called right after given entity e will be removed from World
func (w *World) OnEntityRemove(e tl.Drawable, f func()) {
	w.onEntityRemove[e] = f
}

// IsLowColor is helper function for quick access to LowColor world option in Draw methods
func IsLowColor(s *tl.Screen) bool {
	if world, ok := s.Level().(*World); ok {
		return world.options.LowColor
	}
	panic("Invalid level. World expected")
}

// ZIndexer if implemented by entity allows to be sorted by z-index.
// Lower z-index will be drawn first.
// Zero is used as default value for entities do not implementing ZIndexer.
type ZIndexer interface {
	ZIndex() int
}

// ExtendedLevel extends termloop.Level with additional functionality
type ExtendedLevel interface {
	tl.Level
	OnEntityRemove(e tl.Drawable, f func())
}
